@@ -154,7 +154,9 @@ _GENERIC_CONFIG_OPTIONAL = [
     "CONFIG_UPROBES=y",
     "CONFIG_UPROBE_EVENTS=y",
     "CONFIG_DEBUG_FS=y",
-    "CONFIG_PROVE_LOCKING=y",
+    "# Required to generate memory dumps for drgn",
+    "CONFIG_FW_CFG_SYSFS=y",
+    "CONFIG_FW_CFG_SYSFS_CMDLINE=y",
     "# Graphics support",
     "CONFIG_DRM=y",
     "CONFIG_DRM_VIRTIO_GPU=y",
