@@ -19,6 +19,7 @@ import re
 import itertools
 import subprocess
 import signal
+import termios
 from shutil import which
 from time import sleep
 from base64 import b64encode
@@ -28,7 +29,7 @@ from .. import mkinitramfs
 from .. import qemu_helpers
 from .. import architectures
 from .. import resources
-from ..util import SilentError, uname, get_username
+from ..util import SilentError, uname, get_username, find_binary_or_raise
 
 
 def make_parser() -> argparse.ArgumentParser:
@@ -150,6 +151,12 @@ def make_parser() -> argparse.ArgumentParser:
         default=None,
         help="Set guest memory and qemu -m flag.",
     )
+    g.add_argument(
+        "--numa",
+        action="append",
+        default=None,
+        help="Create NUMA nodes in the guest.",
+    )
     g.add_argument(
         "--cpus", action="store", default=None, help="Set guest cpu and qemu -smp flag."
     )
@@ -226,6 +233,11 @@ def make_parser() -> argparse.ArgumentParser:
         action="store_true",
         help='Avoid using the "microvm" QEMU architecture (only on x86_64)',
     )
+    g.add_argument(
+        "--disable-kvm",
+        action="store_true",
+        help='Avoid using hardware virtualization / KVM',
+    )
     g.add_argument(
         "--force-initramfs",
         action="store_true",
@@ -334,7 +346,12 @@ class Kernel:
 
 
 def get_rootfs_from_kernel_path(path):
-    while path != "/" and not os.path.exists(path + "/lib/modules"):
+    while path and path != "/" and not os.path.exists(path + "/lib/modules"):
+        path, _ = os.path.split(path)
+    # If a distro, like openSUSE Tumbleweed, has /lib symlinked to /usr/lib,
+    # the rootfs may be mistakenly identified as /usr. In such cases, ensure to
+    # get the rootfs from one level higher.
+    if path.endswith("/usr"):
         path, _ = os.path.split(path)
     return os.path.abspath(path)
 
@@ -346,12 +363,20 @@ def get_kernel_version(path):
         )
     if not os.access(path, os.R_OK):
         arg_fail("unable to access %s (check for read permissions)" % path)
-    result = subprocess.run(["file", path], capture_output=True, text=True, check=False)
-    for item in result.stdout.split(", "):
-        match = re.search(r"^[vV]ersion (\S+)", item)
-        if match:
-            kernel_version = match.group(1)
-            return kernel_version
+    try:
+        result = subprocess.run(
+            ["file", path], capture_output=True, text=True, check=False
+        )
+        for item in result.stdout.split(", "):
+            match = re.search(r"^[vV]ersion (\S+)", item)
+            if match:
+                kernel_version = match.group(1)
+                return kernel_version
+    except FileNotFoundError:
+        sys.stderr.write(
+            "warning: `file` is not installed in the system, "
+            "virtme-ng may fail to detect kernel version\n"
+        )
     # 'file' failed to get kernel version, try with 'strings'.
     result = subprocess.run(
         ["strings", path], capture_output=True, text=True, check=False
@@ -365,6 +390,7 @@ def get_kernel_version(path):
 
 def find_kernel_and_mods(arch, args) -> Kernel:
     kernel = Kernel()
+    kernel.config = None
 
     kernel.use_root_mods = False
     if args.installed_kernel is not None:
@@ -416,7 +442,10 @@ def find_kernel_and_mods(arch, args) -> Kernel:
         else:
             # Try to automatically detect modules' path
             root_dir = get_rootfs_from_kernel_path(kernel.kimg)
-            if root_dir == "/":
+            # If we are using the entire host filesystem or if we are using
+            # a chroot (via --root) we don't have to do anything special action
+            # the modules, just rely on /lib/modules in the target rootfs.
+            if root_dir == "/" or args.root != '/':
                 kernel.use_root_mods = True
             elif root_dir.startswith("/tmp"):
                 sys.stderr.write(
@@ -430,11 +459,13 @@ def find_kernel_and_mods(arch, args) -> Kernel:
             else:
                 mod_file = os.path.join(kernel.moddir, "modules.dep")
                 if not os.path.exists(mod_file):
+                    depmod = find_binary_or_raise(["depmod"])
+
                     # Try to refresh modules directory. Some packages (e.g., debs)
                     # don't ship all the required modules information, so we
                     # need to refresh the modules directory using depmod.
                     subprocess.call(
-                        ["depmod", "-a", "-b", root_dir, kver],
+                        [depmod, "-a", "-b", root_dir, kver],
                         stderr=subprocess.DEVNULL,
                     )
                 kernel.modfiles = modfinder.find_modules_from_install(
@@ -469,7 +500,7 @@ def find_kernel_and_mods(arch, args) -> Kernel:
             # Check if modules.order exists, otherwise fallback to mods=none
             if os.path.exists(mod_file):
                 # Check if virtme's kernel modules directory needs to be updated
-                if not os.path.exists(virtme_mods) or is_file_more_recent(
+                if not os.path.exists(virtme_mod_file) or is_file_more_recent(
                     mod_file, virtme_mod_file
                 ):
                     if modmode == "use":
@@ -586,8 +617,13 @@ class VirtioFS:
 
         # Export the whole root fs of the host, do not enable sandbox, otherwise we
         # would get permission errors.
+        if not verbose:
+            stderr = "2>/dev/null"
+        else:
+            stderr = ""
         os.system(
-            f"{virtiofsd_path} --syslog --socket-path {self.sock} --shared-dir {path} --sandbox none &"
+            f"{virtiofsd_path} --syslog --no-announce-submounts " +
+            f"--socket-path {self.sock} --shared-dir {path} --sandbox none {stderr} &"
         )
         max_attempts = 5
         check_duration = 0.1
@@ -599,7 +635,8 @@ class VirtioFS:
             sleep(check_duration)
             check_duration *= 2
         else:
-            print("virtme-run: failed to start virtiofsd, fallback to 9p")
+            if verbose:
+                sys.stderr.write("virtme-run: failed to start virtiofsd, fallback to 9p")
             return False
         return True
 
@@ -624,11 +661,14 @@ def export_virtiofs(
 
     # Adjust qemu options to use virtiofsd
     fsid = "virtfs%d" % len(qemuargs)
-    if memory is None:
-        memory = "128M"
+
     vhost_dev_type = arch.vhost_dev_type()
     qemuargs.extend(["-chardev", f"socket,id=char{fsid},path={virtio_fs.sock}"])
     qemuargs.extend(["-device", f"{vhost_dev_type},chardev=char{fsid},tag={mount_tag}"])
+    if memory is None:
+        memory = "128M"
+    elif memory == 0:
+        return True
     qemuargs.extend(["-object", f"memory-backend-memfd,id=mem,size={memory},share=on"])
     qemuargs.extend(["-numa", "node,memdev=mem"])
 
@@ -692,8 +732,26 @@ def sanitize_disk_args(func: str, arg: str) -> Tuple[str, str]:
     return name, fn
 
 
+def can_access_file(path):
+    if not os.path.exists(path):
+        return False
+    try:
+        fd = os.open(path, os.O_RDWR | os.O_CLOEXEC)
+        os.close(fd)
+        return True
+
+    except OSError:
+        return False
+
+
+def can_use_kvm(args):
+    if args.disable_kvm:
+        return False
+    return can_access_file("/dev/kvm")
+
+
 def can_use_microvm(args):
-    return not args.disable_microvm and args.arch == "x86_64"
+    return not args.disable_microvm and not args.numa and args.arch == "x86_64" and can_use_kvm(args)
 
 
 def has_read_acl(username, file_path):
@@ -717,6 +775,16 @@ def has_read_acl(username, file_path):
         return False  # Error occurred while executing getfacl command
 
 
+def is_statically_linked(binary_path):
+    try:
+        # Run the 'file' command on the binary and check for the string
+        # "statically linked"
+        result = subprocess.check_output(["file", binary_path], universal_newlines=True)
+        return "statically linked" in result
+    except subprocess.CalledProcessError:
+        return False
+
+
 # Allowed characters in mount paths.  We can extend this over time if needed.
 _SAFE_PATH_PATTERN = "[a-zA-Z0-9_+ /.-]+"
 _RWDIR_RE = re.compile("^(%s)(?:=(%s))?$" % (_SAFE_PATH_PATTERN, _SAFE_PATH_PATTERN))
@@ -758,6 +826,23 @@ def do_it() -> int:
             args.memory += "M"
         qemuargs.extend(["-m", args.memory])
 
+    # Propagate /proc/sys/fs/nr_open from the host to the guest, otherwise we
+    # may see some EPERM errors, because certain applications/settings may
+    # expect to be able to use a higher limit of the max number of open files.
+    with open('/proc/sys/fs/nr_open', 'r', encoding="utf-8") as file:
+        nr_open = file.readline().strip()
+        kernelargs.append(f"nr_open={nr_open}")
+
+    # Parse NUMA settings.
+    if args.numa:
+        for i, numa in enumerate(args.numa, start=1):
+            size, cpus = numa.split(",", 1) if "," in numa else (numa, None)
+            cpus = f",{cpus}" if cpus else ""
+            qemuargs.extend([
+                "-object", f"memory-backend-memfd,id=mem{i},size={size},share=on",
+                "-numa", f"node,memdev=mem{i}{cpus}"
+            ])
+
     if args.snaps:
         if args.root == "/":
             snapd_state = "/var/lib/snapd/state.json"
@@ -812,7 +897,10 @@ def do_it() -> int:
             args.root,
             "ROOTFS",
             guest_tools_path=guest_tools_path,
-            memory=args.memory,
+            # virtiofsd requires a NUMA not, if --numa is specified simply use
+            # the user-defined NUMA node, otherwise create a NUMA node with all
+            # the memory.
+            memory=0 if args.numa else args.memory,
             verbose=args.verbose,
         )
         if can_use_microvm(args) and use_virtiofs:
@@ -842,13 +930,15 @@ def do_it() -> int:
             "init=/bin/sh",
             "--",
             "-c",
-            ";".join([
-                "mount -t tmpfs run /run",
-                "mkdir -p /run/virtme/guesttools",
-                "/bin/mount -n -t 9p -o ro,version=9p2000.L,trans=virtio,access=any " +
-                "virtme.guesttools /run/virtme/guesttools",
-                f"exec /run/virtme/guesttools/{virtme_init_cmd}",
-            ]),
+            ";".join(
+                [
+                    "mount -t tmpfs run /run",
+                    "mkdir -p /run/virtme/guesttools",
+                    "/bin/mount -n -t 9p -o ro,version=9p2000.L,trans=virtio,access=any "
+                    + "virtme.guesttools /run/virtme/guesttools",
+                    f"exec /run/virtme/guesttools/{virtme_init_cmd}",
+                ]
+            ),
         ]
 
     # Arrange for modules to end up in the right place
@@ -898,33 +988,49 @@ def do_it() -> int:
         kernelargs.append("virtme_rw_overlay%d=%s" % (i, d))
 
     # Turn on KVM if available
-    if is_native:
+    kvm_ok = can_use_kvm(args)
+    if is_native and kvm_ok:
         qemuargs.extend(["-machine", "accel=kvm:tcg"])
 
     # Add architecture-specific options
-    qemuargs.extend(arch.qemuargs(is_native))
+    qemuargs.extend(arch.qemuargs(is_native, kvm_ok))
 
     # Set up / override baseline devices
     qemuargs.extend(["-parallel", "none"])
     qemuargs.extend(["-net", "none"])
 
     if args.graphics is None and not args.script_sh and not args.script_exec:
-        # It would be nice to use virtconsole, but it's terminally broken
-        # in current kernels.  Nonetheless, I'm configuring the console
-        # manually to make it easier to tweak in the future.
         qemuargs.extend(["-echr", "1"])
         qemuargs.extend(["-serial", "none"])
-        qemuargs.extend(["-chardev", "stdio,id=console,signal=off,mux=on"])
-
-        qemuargs.extend(arch.qemu_serial_console_args())
-
-        qemuargs.extend(["-mon", "chardev=console"])
 
         if args.verbose:
-            kernelargs.extend(arch.earlyconsole_args())
-        qemuargs.extend(arch.qemu_nodisplay_args())
+            # Check if we have permission to access the current stderr.
+            if not can_access_file("/proc/self/fd/2"):
+                print(
+                    "ERROR: not a valid pts, try to run vng inside tmux or screen",
+                    file=sys.stderr,
+                )
+                sys.exit(1)
 
-        kernelargs.extend(arch.serial_console_args())
+            # Redirect kernel messages to stderr, creating a separate console
+            qemuargs.extend(["-chardev", "file,path=/proc/self/fd/2,id=dmesg"])
+            qemuargs.extend(["-device", arch.virtio_dev_type("serial")])
+            qemuargs.extend(["-device", "virtconsole,chardev=dmesg"])
+            kernelargs.extend(["console=hvc0"])
+
+            # Unfortunately we can't use hvc0 to redirect early console
+            # messages to stderr, so just send them to the main console, in
+            # this way we don't lose early printk's in verbose mode and we can
+            # catch potential boot issues.
+            kernelargs.extend(arch.earlyconsole_args())
+
+        qemuargs.extend(["-chardev", "stdio,id=console,signal=off,mux=on"])
+        qemuargs.extend(["-serial", "chardev:console"])
+        qemuargs.extend(["-mon", "chardev=console"])
+
+        kernelargs.extend(["virtme_console=" + arg for arg in arch.serial_console_args()])
+
+        qemuargs.extend(arch.qemu_nodisplay_args())
 
         # PS/2 probing is slow; give the kernel a hint to speed it up.
         kernelargs.extend(["psmouse.proto=exps"])
@@ -989,12 +1095,37 @@ def do_it() -> int:
                 ]
             )
 
-    def do_script(shellcmd: str, show_boot_console=False) -> None:
-        if args.graphics is not None:
-            arg_fail("scripts and --graphics are mutually exclusive")
+    ret_path = None
 
-        # Turn off default I/O
-        qemuargs.extend(arch.qemu_nodisplay_args())
+    def cleanup_script_retcode():
+        os.unlink(ret_path)
+
+    def fetch_script_retcode():
+        if ret_path is None:
+            return None
+        try:
+            with open(ret_path, 'r', encoding="utf-8") as file:
+                number_str = file.read().strip()
+                if number_str.isdigit():
+                    return int(number_str)
+                return None
+        except FileNotFoundError:
+            return None
+
+    def do_script(shellcmd: str, ret_path=None, show_boot_console=False) -> None:
+        if args.graphics is None:
+            # Turn off default I/O
+            qemuargs.extend(arch.qemu_nodisplay_args())
+
+        # Check if we can redirect stdin/stdout/stderr.
+        if not can_access_file("/proc/self/fd/0") or \
+           not can_access_file("/proc/self/fd/1") or \
+           not can_access_file("/proc/self/fd/2"):
+            print(
+                "ERROR: not a valid pts, try to run vng inside tmux or screen",
+                file=sys.stderr,
+            )
+            sys.exit(1)
 
         # Configure kernel console output
         if show_boot_console:
@@ -1006,7 +1137,7 @@ def do_it() -> int:
         qemuargs.extend(arch.qemu_serial_console_args())
         qemuargs.extend(["-chardev", f"file,id=console,path={output}"])
 
-        kernelargs.extend(arch.serial_console_args())
+        kernelargs.extend(["console=" + arg for arg in arch.serial_console_args()])
         kernelargs.extend(arch.earlyconsole_args())
         kernelargs.extend(console_args)
 
@@ -1043,15 +1174,17 @@ def do_it() -> int:
             ["-device", "virtserialport,name=virtme.dev_stderr,chardev=dev_stderr"]
         )
 
-        # Scripts shouldn't reboot
+        # Create a virtio serial device to channel the retcode of the script
+        # executed in the guest to the host.
+        if ret_path is not None:
+            qemuargs.extend(["-chardev", f"file,id=ret,path={ret_path}"])
+            qemuargs.extend(["-device", arch.virtio_dev_type("serial")])
+            qemuargs.extend(["-device", "virtserialport,name=virtme.ret,chardev=ret"])
+
+        # Scripts shouldn't reboot and shouldn't hang on panic: make sure to
+        # force an exit condition if a panic happens.
         qemuargs.extend(["-no-reboot"])
-
-        # Encode the shell command to base64 to handle special characters (such
-        # as quotes, double quotes, etc.).
-        shellcmd = b64encode(shellcmd.encode('utf-8')).decode('utf-8')
-
-        # Ask virtme-init to run the script
-        kernelargs.append(f"virtme.exec=`{shellcmd}`")
+        kernelargs.append("panic=-1")
 
         # Nasty issue: QEMU will set O_NONBLOCK on fds 0, 1, and 2.
         # This isn't inherently bad, but it can cause a problem if
@@ -1070,8 +1203,24 @@ def do_it() -> int:
                     os.dup2(newfd, oldfd)
                     os.close(newfd)
 
+        # Encode the shell command to base64 to handle special characters (such
+        # as quotes, double quotes, etc.).
+        shellcmd = b64encode(shellcmd.encode("utf-8")).decode("utf-8")
+
+        if args.graphics is not None:
+            kernelargs.append("virtme_graphics=1")
+
+        # Ask virtme-init to run the script
+        kernelargs.append(f"virtme.exec=`{shellcmd}`")
+
+    # Do not break old syntax "-g command"
+    if args.graphics is not None and args.script_sh is None:
+        args.script_sh = args.graphics
+
     if args.script_sh is not None:
-        do_script(args.script_sh, show_boot_console=args.show_boot_console)
+        _, ret_path = tempfile.mkstemp(prefix="virtme_ret")
+        atexit.register(cleanup_script_retcode)
+        do_script(args.script_sh, ret_path=ret_path, show_boot_console=args.show_boot_console)
 
     if args.script_exec is not None:
         do_script(
@@ -1083,8 +1232,6 @@ def do_it() -> int:
         video_args = arch.qemu_display_args()
         if video_args:
             qemuargs.extend(video_args)
-        if args.graphics != "":
-            kernelargs.append("virtme_graphics=%s" % args.graphics)
 
     if args.net:
         qemuargs.extend(["-device", "%s,netdev=n0" % arch.virtio_dev_type("net")])
@@ -1122,6 +1269,12 @@ def do_it() -> int:
     if args.user:
         kernelargs.append("virtme_user=%s" % args.user)
 
+    # If we are running as root on the host pass this information to the guest
+    # (this can be useful to properly support running virtme-ng instances
+    # inside docker)
+    if os.geteuid() == 0:
+        kernelargs.append("virtme_root_user=1")
+
     initrdpath: Optional[str]
 
     if need_initramfs:
@@ -1135,6 +1288,13 @@ def do_it() -> int:
                     file=sys.stderr,
                 )
                 return 1
+            if not is_statically_linked(busybox):
+                print(
+                    "virtme-run: a statically linked busybox could not be found, "
+                    "please install busybox-static",
+                    file=sys.stderr,
+                )
+                return 1
             config.busybox = busybox
 
         if args.rw:
@@ -1224,7 +1384,15 @@ def do_it() -> int:
         if pid:
             try:
                 pid, status = os.waitpid(pid, 0)
-                return status
+                ret = fetch_script_retcode()
+                if ret is not None:
+                    return ret
+                if not args.script_sh and not args.script_exec:
+                    return status
+                # Return special error code 255 in case of unexpected exit
+                # (e.g., kernel panic).
+                return 255
+
             except KeyboardInterrupt:
                 sys.stderr.write("Interrupted.")
                 sys.exit(1)
@@ -1233,11 +1401,30 @@ def do_it() -> int:
     return 0
 
 
+def save_terminal_settings():
+    return termios.tcgetattr(sys.stdin) if sys.stdin.isatty() else None
+
+
+def restore_terminal_settings(settings):
+    if settings is not None:
+        termios.tcsetattr(sys.stdin, termios.TCSAFLUSH, settings)
+
+
+def signal_handler(_signum, _frame):
+    sys.exit(1)
+
+
 def main() -> int:
+    # Catch potential signals that may interrupt the execution (SIGTERM) and
+    # make sure the terminal settings are restored on exit.
+    settings = save_terminal_settings()
+    signal.signal(signal.SIGTERM, signal_handler)
     try:
         return do_it()
     except SilentError:
         return 1
+    finally:
+        restore_terminal_settings(settings)
 
 
 if __name__ == "__main__":
