@@ -4,6 +4,7 @@
 """virtme-ng: main command-line frontend."""
 
 import argparse
+import re
 import os
 import sys
 import socket
@@ -20,26 +21,19 @@ from subprocess import (
 )
 from select import select
 from pathlib import Path
-from argcomplete import autocomplete
+try:
+    from argcomplete import autocomplete
+except ModuleNotFoundError:
+    def autocomplete(*args, **kwargs):
+        # pylint: disable=unused-argument
+        pass
 
 from virtme.util import SilentError, uname, get_username
-from virtme_ng.utils import CONF_FILE
-from virtme_ng.spinner import Spinner
+from virtme_ng.utils import CONF_FILE, spinner_decorator
+from virtme_ng.mainline import KernelDownloader
 from virtme_ng.version import VERSION
 
 
-def spinner_decorator(message):
-    def decorator(func):
-        def wrapper(*args, **kwargs):
-            with Spinner(message=message):
-                result = func(*args, **kwargs)
-                return result
-
-        return wrapper
-
-    return decorator
-
-
 def check_call_cmd(command, quiet=False, dry_run=False):
     if dry_run:
         print(" ".join(command))
@@ -98,8 +92,13 @@ def make_parser():
         nargs="?",
         const=uname.release,
         default=None,
-        help="Run specified kernel image or an installed kernel version. "
-        "If no argument is specified the running kernel will be used.",
+        help="Run a specified kernel; "
+        "--run can accept one of the following arguments: 1) nothing (in this "
+        "case it'll try to boot the same kernel running on the host), 2) a kernel "
+        "binary (like ./arch/x86/boot/bzImage), 3) a directory (where it'll try "
+        "to find a valid kernel binary file), 4) an upstream version, for "
+        "example `vng --run v6.6.17` (in this case vng will download a "
+        "precompiled upstream kernel from the Ubuntu mainline repository)",
     )
 
     g_action.add_argument(
@@ -135,7 +134,7 @@ def make_parser():
         "--skip-config",
         "-s",
         action="store_true",
-        help="Do not re-generate kernel .config",
+        help="[deprecated] Do not re-generate kernel .config",
     )
 
     parser.add_argument(
@@ -200,7 +199,10 @@ def make_parser():
     parser.add_argument("--qemu", action="store", help="Use the specified QEMU binary")
 
     parser.add_argument(
-        "--name", action="store", help="Set guest hostname and qemu -name flag"
+        "--name",
+        action="store",
+        default="virtme-ng",
+        help="Set guest hostname and qemu -name flag"
     )
 
     parser.add_argument(
@@ -216,6 +218,12 @@ def make_parser():
         + "(useful with --arch)",
     )
 
+    parser.add_argument(
+        "--root-release",
+        action="store",
+        help="Use a target Ubuntu release to create a new chroot (used with --root)",
+    )
+
     parser.add_argument(
         "--rw",
         action="store_true",
@@ -233,6 +241,12 @@ def make_parser():
         help='Avoid using the "microvm" QEMU architecture (only on x86_64)',
     )
 
+    parser.add_argument(
+        "--disable-kvm",
+        action="store_true",
+        help='Avoid using hardware virtualization / KVM',
+    )
+
     parser.add_argument(
         "--cwd",
         action="store",
@@ -278,6 +292,16 @@ def make_parser():
         "--memory", "-m", action="store", help="Set guest memory size (qemu -m flag)"
     )
 
+    parser.add_argument(
+        "--numa",
+        metavar="MEM[,cpus=FIRST_CPU1[-LAST_CPU1]][,cpus=FIRST_CPU2[-LAST_CPU2]]...",
+        action="append",
+        help="Create a NUMA node in the guest. "
+        + "Use this option multiple times to create more NUMA nodes. "
+        + "The total memory size assigned to NUMA nodes must match the guest memory size (specified with --memory/-m). "
+        + "This option implicitly disables the microvm architecture."
+    )
+
     parser.add_argument(
         "--balloon",
         action="store_true",
@@ -329,12 +353,8 @@ def make_parser():
     parser.add_argument(
         "--graphics",
         "-g",
-        action="store",
-        nargs="?",
-        const="",
-        metavar="BINARY",
-        help="Show graphical output instead of using a console. "
-        + "An argument can be optionally specified to start a graphical application.",
+        action="store_true",
+        help="Show graphical output instead of using a console.",
     )
 
     parser.add_argument(
@@ -344,6 +364,13 @@ def make_parser():
         help="Increase console output verbosity.",
     )
 
+    parser.add_argument(
+        "--quiet",
+        "-q",
+        action="store_true",
+        help="Override verbose mode (disable --verbose).",
+    )
+
     parser.add_argument(
         "--qemu-opts",
         "-o",
@@ -454,17 +481,18 @@ git reset --hard __virtme__
 """
 
 
-def create_root(destdir, arch):
+def create_root(destdir, arch, release):
     """Initialize a rootfs directory, populating files/directory if it doesn't exist."""
     if os.path.exists(destdir):
         return
     # Use Ubuntu's cloud images to create a rootfs, these images are fairly
     # small and they provide a nice environment to test kernels.
-    release = (
-        check_output("lsb_release -s -c", shell=True)
-        .decode(sys.stdout.encoding)
-        .rstrip()
-    )
+    if release is None:
+        release = (
+            check_output("lsb_release -s -c", shell=True)
+            .decode(sys.stdout.encoding)
+            .rstrip()
+        )
     url = (
         "https://cloud-images.ubuntu.com/"
         + f"{release}/current/{release}-server-cloudimg-{arch}-root.tar.xz"
@@ -693,6 +721,30 @@ class KernelSource:
         else:
             self.virtme_param["name"] = "--name " + socket.gethostname()
 
+    def _get_virtme_exec(self, args):
+        if args.envs:
+            args.exec = " ".join(args.envs)
+        if args.exec is not None:
+            self.virtme_param["exec"] = f'--script-sh "{args.exec}"'
+        else:
+            self.virtme_param["exec"] = ""
+
+    def _get_virtme_user(self, args):
+        # Default user for scripts is root, default user for interactive
+        # sessions is current user.
+        #
+        # NOTE: graphic sessions are considered interactive.
+        if args.exec and not args.graphics:
+            self.virtme_param["user"] = ""
+        else:
+            self.virtme_param["user"] = "--user " + get_username()
+        # Override default user, if specified by the --user argument.
+        if args.user is not None:
+            if args.user != "root":
+                self.virtme_param["user"] = "--user " + args.user
+            else:
+                self.virtme_param["user"] = ""
+
     def _get_virtme_arch(self, args):
         if args.arch is not None:
             if args.arch not in ARCH_MAPPING:
@@ -703,17 +755,9 @@ class KernelSource:
         else:
             self.virtme_param["arch"] = ""
 
-    def _get_virtme_user(self, args):
-        if args.user is not None:
-            self.virtme_param["user"] = "--user " + args.user
-        elif args.root is None:
-            self.virtme_param["user"] = "--user " + get_username()
-        else:
-            self.virtme_param["user"] = ""
-
     def _get_virtme_root(self, args):
         if args.root is not None:
-            create_root(args.root, args.arch or "amd64")
+            create_root(args.root, args.arch or "amd64", args.root_release)
             self.virtme_param["root"] = f"--root {args.root}"
         else:
             self.virtme_param["root"] = ""
@@ -746,18 +790,35 @@ class KernelSource:
 
     def _get_virtme_overlay_rwdir(self, args):
         # Set default overlays if rootfs is mounted in read-only mode.
-        if not args.rw:
+        if args.rw:
+            self.virtme_param["overlay_rwdir"] = ""
+        else:
             self.virtme_param["overlay_rwdir"] = " ".join(
                 f"--overlay-rwdir {d}"
-                for d in ("/etc", "/home", "/opt", "/srv", "/usr", "/var")
+                for d in ("/etc", "/lib", "/home", "/opt", "/srv", "/usr", "/var")
             )
         # Add user-specified overlays.
         for item in args.overlay_rwdir:
-            self.virtme_param["overlay_rwdir"] += "--overlay-rwdir " + item
+            self.virtme_param["overlay_rwdir"] += " --overlay-rwdir " + item
 
     def _get_virtme_run(self, args):
         if args.run is not None:
-            self.virtme_param["kdir"] = "--kimg " + args.run
+            # If an upstream version is specified (using an upstream tag) fetch
+            # and run the corresponding kernel from the Ubuntu mainline
+            # repository.
+            if re.match(r'^v\d+(\.\d+)*$', args.run):
+                if args.arch is None:
+                    arch = 'amd64'
+                else:
+                    arch = args.arch
+                try:
+                    mainline = KernelDownloader(args.run, arch=arch, verbose=args.verbose)
+                    self.virtme_param["kdir"] = "--kimg " + mainline.target
+                except FileNotFoundError as exc:
+                    sys.stderr.write(str(exc) + "\n")
+                    sys.exit(1)
+            else:
+                self.virtme_param["kdir"] = "--kimg " + args.run
         else:
             self.virtme_param["kdir"] = "--kdir ./"
 
@@ -779,12 +840,6 @@ class KernelSource:
         else:
             self.virtme_param["no_virtme_ng_init"] = ""
 
-    def _get_virtme_exec(self, args):
-        if args.exec is not None:
-            self.virtme_param["exec"] = f'--script-sh "{args.exec}"'
-        else:
-            self.virtme_param["exec"] = ""
-
     def _get_virtme_network(self, args):
         if args.network is not None:
             self.virtme_param["network"] = f"--net {args.network}"
@@ -807,11 +862,19 @@ class KernelSource:
             self.virtme_param["sound"] = ""
 
     def _get_virtme_disable_microvm(self, args):
-        if args.disable_microvm:
+        # Automatically disable microvm in debug mode, since it seems to
+        # produce incomplete memory dumps.
+        if args.disable_microvm or args.debug:
             self.virtme_param["disable_microvm"] = "--disable-microvm"
         else:
             self.virtme_param["disable_microvm"] = ""
 
+    def _get_virtme_disable_kvm(self, args):
+        if args.disable_kvm:
+            self.virtme_param["disable_kvm"] = "--disable-kvm"
+        else:
+            self.virtme_param["disable_kvm"] = ""
+
     def _get_virtme_9p(self, args):
         if args.force_9p:
             self.virtme_param["force_9p"] = "--force-9p"
@@ -825,8 +888,8 @@ class KernelSource:
             self.virtme_param["force_initramfs"] = ""
 
     def _get_virtme_graphics(self, args):
-        if args.graphics is not None:
-            self.virtme_param["graphics"] = f'--graphics "{args.graphics}"'
+        if args.graphics:
+            self.virtme_param["graphics"] = '--graphics'
         else:
             self.virtme_param["graphics"] = ""
 
@@ -853,6 +916,13 @@ class KernelSource:
         else:
             self.virtme_param["memory"] = "--memory " + args.memory
 
+    def _get_virtme_numa(self, args):
+        if args.numa is not None:
+            numa_str = " ".join([f"--numa {numa}" for numa in args.numa])
+            self.virtme_param["numa"] = numa_str
+        else:
+            self.virtme_param["numa"] = ""
+
     def _get_virtme_balloon(self, args):
         if args.balloon:
             self.virtme_param["balloon"] = "--balloon"
@@ -889,7 +959,10 @@ class KernelSource:
         if args.qemu_opts is not None:
             qemu_args += " ".join(args.qemu_opts)
         if args.debug:
-            qemu_args += "-s -qmp tcp:localhost:3636,server,nowait"
+            # Enable vmcoreinfo (required by drgn memory dumps)
+            qemu_args += "-device vmcoreinfo "
+            # Enable debug mode and QMP (to trigger memory dump via `vng --dump`)
+            qemu_args += "-s -qmp tcp:localhost:3636,server,nowait "
         if qemu_args != "":
             self.virtme_param["qemu_opts"] = "--qemu-opts " + qemu_args
         else:
@@ -898,6 +971,7 @@ class KernelSource:
     def run(self, args):
         """Execute a kernel inside virtme-ng."""
         self._get_virtme_name(args)
+        self._get_virtme_exec(args)
         self._get_virtme_user(args)
         self._get_virtme_arch(args)
         self._get_virtme_root(args)
@@ -910,11 +984,11 @@ class KernelSource:
         self._get_virtme_dry_run(args)
         self._get_virtme_no_virtme_ng_init(args)
         self._get_virtme_mods(args)
-        self._get_virtme_exec(args)
         self._get_virtme_network(args)
         self._get_virtme_disk(args)
         self._get_virtme_sound(args)
         self._get_virtme_disable_microvm(args)
+        self._get_virtme_disable_kvm(args)
         self._get_virtme_9p(args)
         self._get_virtme_initramfs(args)
         self._get_virtme_graphics(args)
@@ -922,6 +996,7 @@ class KernelSource:
         self._get_virtme_append(args)
         self._get_virtme_cpus(args)
         self._get_virtme_memory(args)
+        self._get_virtme_numa(args)
         self._get_virtme_balloon(args)
         self._get_virtme_snaps(args)
         self._get_virtme_busybox(args)
@@ -932,6 +1007,7 @@ class KernelSource:
         cmd = (
             "virtme-run "
             + f'{self.virtme_param["name"]} '
+            + f'{self.virtme_param["exec"]} '
             + f'{self.virtme_param["user"]} '
             + f'{self.virtme_param["arch"]} '
             + f'{self.virtme_param["root"]} '
@@ -944,11 +1020,11 @@ class KernelSource:
             + f'{self.virtme_param["dry_run"]} '
             + f'{self.virtme_param["no_virtme_ng_init"]} '
             + f'{self.virtme_param["mods"]} '
-            + f'{self.virtme_param["exec"]} '
             + f'{self.virtme_param["network"]} '
             + f'{self.virtme_param["disk"]} '
             + f'{self.virtme_param["sound"]} '
             + f'{self.virtme_param["disable_microvm"]} '
+            + f'{self.virtme_param["disable_kvm"]} '
             + f'{self.virtme_param["force_9p"]} '
             + f'{self.virtme_param["force_initramfs"]} '
             + f'{self.virtme_param["graphics"]} '
@@ -956,6 +1032,7 @@ class KernelSource:
             + f'{self.virtme_param["append"]} '
             + f'{self.virtme_param["cpus"]} '
             + f'{self.virtme_param["memory"]} '
+            + f'{self.virtme_param["numa"]} '
             + f'{self.virtme_param["balloon"]} '
             + f'{self.virtme_param["snaps"]} '
             + f'{self.virtme_param["busybox"]} '
@@ -966,11 +1043,6 @@ class KernelSource:
 
     def dump(self, args):
         """Generate or analyze a crash memory dump."""
-        if not os.path.isfile("vmlinux"):
-            arg_fail(
-                "vmlinux not found, try to recompile the kernel with "
-                + "--build-host-vmlinux (if --build-host was used)"
-            )
         # Use QMP to generate a memory dump
         sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
         sock.connect(("localhost", 3636))
@@ -989,7 +1061,7 @@ class KernelSource:
         with tempfile.NamedTemporaryFile(delete=dump_file is None) as tmp:
             msg = (
                 '{"execute":"dump-guest-memory",'
-                '"arguments":{"paging":false,'
+                '"arguments":{"paging":true,'
                 '"protocol":"file:' + tmp.name + '"}}'
                 "\r"
             )
@@ -1050,8 +1122,7 @@ def clean(kern_source, args):
 
 def run(kern_source, args):
     """Run the kernel."""
-    kern_source.run(args)
-    return True
+    return kern_source.run(args)
 
 
 @spinner_decorator(message="🐞 generating memory dump")
@@ -1071,23 +1142,30 @@ def do_it() -> int:
         for opt in kern_source.default_opts:
             val = kern_source.default_opts[opt]
             setattr(args, opt, val)
+
+    if args.verbose and args.quiet:
+        args.verbose = False
     try:
         if args.clean:
             clean(kern_source, args)
         elif args.dump is not None:
             dump(kern_source, args)
+        elif args.build or args.kconfig:
+            if args.commit:
+                checkout(kern_source, args)
+            config(kern_source, args)
+            if args.kconfig:
+                return 0
+            make(kern_source, args)
         else:
-            if args.build:
-                if args.commit:
-                    checkout(kern_source, args)
-                if not args.skip_config:
-                    config(kern_source, args)
-                    if args.kconfig:
-                        return
-                make(kern_source, args)
-            run(kern_source, args)
+            try:
+                run(kern_source, args)
+                return 0
+            except CalledProcessError as exc:
+                return exc.returncode
     except CalledProcessError as exc:
         raise SilentError() from exc
+    return 0
 
 
 def main() -> int:
